indexing

	description: "Base class for simple CGI programs. Outputs XHTML."

	usage: "Never write to stdout, use `puts' or `add_data' instead."

	known_bugs:
		"1. Does not handle inputs with same name (checkboxes) correctly."

	todo:
		"1. should use binary search allowed_new_line_tag"

	author: "Berend de Boer"
	date: "$Date: 2003/01/09 $"
	revision: "$Revision: #4 $"


deferred class

	EPX_CGI


inherit

	STDC_CURRENT_PROCESS

	EPX_XHTML_WRITER
		rename
			make as make_xhtml_writer
		redefine
			extend,
			extend_character
		end

	EPX_MIME_FIELD_NAMES
		export {NONE} all end

	EXCEPTIONS
		export
			{NONE} all
		end


feature {NONE} -- Initialization

	make is
		do
			make_no_rescue
		rescue
			handle_fatal_error
		end

	make_no_rescue is
		do
			make_xhtml_writer
			debug ("cgi")
				dump_input
				exit (EXIT_FAILURE)
			end
			execute
		end


feature -- the output routine

	execute is
			-- to be implemented by child
		deferred
		end


feature {NONE} -- When script fails

	handle_fatal_error is
			-- Your last change to emit something when the script has
			-- failed unexpectedly. Call only in rescue clause.
		require
			is_exception: exception /= 0
		do
			if not is_header_written then
				content_text_plain
			end
			stdout.puts ("A fatal error prevents this program from continuing.%N")
			if is_developer_exception then
				stdout.puts ("The error message was: ")
				stdout.puts (developer_exception_name)
				stdout.puts ("%N")
			end
			stdout.flush
			exit (EXIT_FAILURE)
		end


feature -- debug support

	dump_input is
			-- Write cgi input to /tmp/cgi_input.
			-- First line contains the content header, is not actually in input!
		local
			file: STDC_BINARY_FILE
			clen: INTEGER
			i: INTEGER
		do
			create file.create_write ("/tmp/cgi_input")
			if request_method.is_equal ("GET") then
				file.puts (query_string)
			elseif request_method.is_equal ("POST") then
				file.write_string (mime_header)
				from
					i := 0
					clen := content_length
				until
					i >= clen
				loop
					stdin.get_character
					file.put_character (stdin.last_byte)
					i := i + 1
				end
			end
			file.close
		end


feature {NONE} -- Overrule some xml stuff

	extend_character (c: CHARACTER) is
			-- Output a character, never write to `stdout' directly!
		do
			stdout.write_character (c)
			precursor (c)
		end

	extend (stuff: STRING) is
			-- Output cgi code, never write to `stdout' directly!
		do
			if stuff /= Void then
				-- Have to do stuff.out here, else postcondition violation
				-- in UC_STRING. To be fixed by adding UC_STRING support
				-- to STDC_FILE.
				stdout.write_string (stuff.out)
			end
			precursor (stuff)
		end


feature -- Standard variables

	auth_type: STRING is
			-- type of authentication used
		do
			Result := env_auth_type.value
		end

	content_type: STRING is
			-- MIME type of data when invoked with POST method
		do
			Result := env_content_type.value
		end

	content_length: INTEGER is
			-- length, in bytes, of data when invoked with POST method
		do
			if env_content_length.value.is_equal ("") then
				Result := 0
			else
				Result := env_content_length.value.to_integer
			end
		end

	gateway_interface: STRING is
			-- name and version of the gateway, for example CGI/1.1
		do
			Result := env_gateway_interface.value
		end

	http_accept: STRING is
			-- contents of the Accept header line sent by the client
		do
			Result := env_http_accept.value
		end

	http_referer: STRING is
			-- contents of the Referer header line
		do
			Result := env_http_referer.value
		end

	http_user_agent: STRING is
			-- name of the client program that is making the request
		do
			Result := env_http_user_agent.value
		end

	path_info: STRING is
			-- extra path information as it was passed to the server in
			-- the query URL
		do
			Result := env_path_info.value
		end

	path_translated: STRING is
			-- extra path information translated to a final, usable
			-- form. The Web document root is prepended to the query
			-- path, and any other path translations are executed.
		do
			Result := env_path_translated.value
		end

	query_string: STRING is
			-- the input when invoked with the GET method
		do
			Result := env_query_string.value
		end

	remote_addr, remote_address: STRING is
			-- IP address of the client that made the request
		do
			Result := env_remote_addr.value
		end

	remote_host: STRING is
			-- name of the remote computer that made the request
		do
			Result := env_remote_host.value
		end

	remote_ident: STRING is
			-- user name as given by the ident protocol
		do
			Result := env_remote_ident.value
		end

	remote_user: STRING is
			-- name of the remote user that made the request
		do
			Result := env_remote_user.value
		end

	request_method: STRING is
			-- name of the method used to invoke the CGI
			-- application. Valid values are GET and POST
		do
			Result := env_request_method.value
		end

	script_name: STRING is
			-- name of script that was invoked
		do
			Result := env_script_name.value
		end

	server_name: STRING is
			-- domain name of the computer that is running the server software
		do
			Result := env_server_name.value
		end

	server_port: INTEGER is
			-- TCP port number on which the server that invoked the CGI
			-- application is operating
		local
			s: STRING
		do
			s := env_server_port.value
			if s.is_empty then
				Result := 0
			else
				Result := s.to_integer
			end
		end

	server_protocol: STRING is
			-- name of the protocol that the server is using and the
			-- version of that protocol. The protocol name and version
			-- are separated by a forward slash with no spaces, for
			-- instance HTTP/1.0
		do
			Result := env_server_protocol.value
		end

	server_software: STRING is
			-- name of the server that is handling the request
		do
			Result := env_server_software.value
		end


feature {NONE} -- Environment variables

	env_auth_type: STDC_ENV_VAR is
		once
			create Result.make ("AUTH_TYPE")
		end

	env_content_length: STDC_ENV_VAR is
		once
			create Result.make ("CONTENT_LENGTH")
		end

	env_content_type: STDC_ENV_VAR is
		once
			create Result.make ("CONTENT_TYPE")
		end

	env_gateway_interface: STDC_ENV_VAR is
		once
			create Result.make ("GATEWAY_INTERFACE")
		end

	env_http_accept: STDC_ENV_VAR is
		once
			create Result.make ("HTTP_ACCEPT")
		end

	env_http_referer: STDC_ENV_VAR is
		once
			create Result.make ("HTTP_REFERER")
		end

	env_http_user_agent: STDC_ENV_VAR is
		once
			create Result.make ("HTTP_USER_AGENT")
		end

	env_path_info: STDC_ENV_VAR is
		once
			create Result.make ("PATH_INFO")
		end

	env_path_translated: STDC_ENV_VAR is
		once
			create Result.make ("PATH_TRANSLATED")
		end

	env_query_string: STDC_ENV_VAR is
		once
			create Result.make ("QUERY_STRING")
		end

	env_remote_addr: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_ADDR")
		end

	env_remote_host: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_HOST")
		end

	env_remote_user: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_USER")
		end

	env_remote_ident: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_IDENT")
		end

	env_request_method: STDC_ENV_VAR is
		once
			create Result.make ("REQUEST_METHOD")
		end

	env_script_name: STDC_ENV_VAR is
		once
			create Result.make ("SCRIPT_NAME")
		end

	env_server_name: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_NAME")
		end

	env_server_port: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_PORT")
		end

	env_server_protocol: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_PROTOCOL")
		end

	env_server_software: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_SOFTWARE")
		end


feature -- Standard cgi header

	content_text is
		obsolete "content_text is obsolete, use context_text_html"
		do
			content_text_html
		end

	content_text_html is
		do
			extend ("Content-Type: text/html")
			new_line
			new_line
			stdout.flush
		end

	content_text_plain is
		do
			extend ("Content-Type: text/plain")
			new_line
			new_line
			stdout.flush
		end


feature -- Server push, multipart header

	content_multipart_x_mixed_replace (boundary: STRING) is
		require
			valid_boundary: boundary /= Void and then not boundary.is_empty
		do
			extend ("Content-Type: multipart/x-mixed-replace;boundary=")
			extend (boundary)
			my_boundary := boundary
			new_line
			new_line
			-- start first part immediately
			extend (my_boundary)
			new_line
			stdout.flush
		end

	content_next_part is
			-- write boundary so next part of multipart msg can be written
		require
			multipart: is_multipart_message
		do
			new_line
			extend (my_boundary)
			new_line
			stdout.flush
		end

	content_multipart_end is
			-- write boundary of multipart
		require
			multipart: is_multipart_message
		do
			new_line
			extend (my_boundary)
			extend ("--")
			new_line
			stdout.flush
		end

	is_multipart_message: BOOLEAN is
		do
			Result := my_boundary /= Void
		end


feature {NONE} -- Private boundary state

	my_boundary: STRING
			-- Used for server push, multipart header.


feature -- Form input

	exist_value (key: STRING): BOOLEAN is
		obsolete "Use has_key instead."
		do
			Result := has_key (key)
		end

	has_input: BOOLEAN is
			-- True if input passed to cgi program.
		do
			Result := not request_method.is_empty
		end

	has_key (key: STRING): BOOLEAN is
			-- Returns True if key defined in input.
		require
			key_not_empty: key /= Void and then not key.is_empty
		do
			Result := cgi_data.has (key)
		end

	is_meta_char (c: CHARACTER): BOOLEAN is
			-- Return True if c is a commonly used meta characters.
		do
			Result := meta_chars.has (c)
		end

	meta_chars: STRING is
		once
			-- Check if this list is correct...
			Result := "`[]<>|&`\%%"
		end

	raw_value (key: STRING): STRING is
			-- Returns value for key.
			-- if key does not exist, the empty string is returned.
		require
			key_not_empty: key /= Void and then not key.is_empty
		do
			search_key (key)
			if found_key then
				create Result.make_from_string (found_key_item.value)
			else
				Result := ""
			end
		ensure
			result_not_void: Result /= Void
		end

	remove_meta_chars (s: STRING) is
			-- If s contains meta characters, they're removed.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > s.count
			loop
				if is_meta_char (s.item (i)) then
					s.remove (i)
				else
					i := i + 1
				end
			end
		end

	value (key: STRING): STRING is
			-- Returns safe value for key, meta characters are removed.
		do
			Result := raw_value (key)
			remove_meta_chars (Result)
		end


feature {NONE} -- Cached key/value

	assert_key_value_pairs_created is
			-- Make sure passed form contents (if any) is parsed.
		do
			if cgi_data = Void then
				create cgi_data.make (32)
				if has_input then
					fill_key_value_pairs
				end
			end
		end

	cgi_data: DS_HASH_TABLE [EPX_KEY_VALUE, STRING]
			-- Array with key-value pairs.

	cgi_input: STRING
			-- Raw input data.

	exist_key: BOOLEAN is
		obsolete "Use found_key instead."
		do
			Result := found_key
		end

	fill_key_value_pairs is
			-- Fill cgi_data, recognizes default and multipart encodings.
		require
			key_value_pair_array_exists: cgi_data /= Void
			there_is_input: has_input
		local
			ctype: STRING
		do
			create ctype.make_from_string (content_type)
			ctype.to_lower
			if ctype.is_empty or else ctype.is_equal (standard_encoding) then
				fill_key_value_pairs_from_standard_encoding
			elseif ctype.substring_index (multipart_encoding, 1) = 1 then
				fill_key_value_pairs_from_multipart_encoding
			end
		end

	found_key: BOOLEAN
			-- Was `search_key' successful?

	found_key_item: EPX_KEY_VALUE
			-- Item found by `search_key'.

	last_pair: EPX_KEY_VALUE is
		obsolete "Use `found_key_item instead."
		do
			Result := found_key_item
		end

	locate_key (key: STRING) is
			-- Locate key in key/value pairs.
			-- last_pair is not Void if found.
		obsolete "Use `search_key' instead"
		do
			search_key (key)
		end

	search_key (key: STRING) is
			-- Locate key in key/value pairs.
			-- last_pair is not Void if found.
		require
			valid_key: key /= Void and then not key.is_empty
		do
			assert_key_value_pairs_created
			cgi_data.search (key)
			if cgi_data.found then
				found_key := True
				found_key_item := cgi_data.found_item
			else
				found_key := False
				found_key_item := Void
			end
		end

	set_cgi_input is
		local
			len: INTEGER
			i: INTEGER
		do
			if request_method.is_equal ("GET") then
				cgi_input := query_string
			elseif request_method.is_equal ("POST") then
				from
					len := content_length
					i := 0
					cgi_input := sh.make_with_capacity (64)
				until
					i = len
				loop
					stdin.read_character
					cgi_input.append_character (stdin.last_character)
					i := i + 1
				end
			end
		end


feature {NONE} -- Standard or multipart key/value filling

	do_fill_key_value_pairs_from_mime (name: STRING; part: EPX_MIME_PART) is
			-- Fill `cgi_data'.
		require
			part_not_void: part /= Void
		local
			sub_part: EPX_MIME_PART
			body: EPX_MIME_BODY
			text_body: EPX_MIME_BODY_TEXT
			file_body: EPX_MIME_BODY_FILE
			field: EPX_MIME_FIELD_WITH_PARAMETERS
			parameter: EPX_MIME_PARAMETER
			keyname,
			keyvalue: STRING
			i: INTEGER
			kv: EPX_KEY_VALUE
			temporary_file: STDC_FILE
		do
			-- `part'.`body' contain the parameters.
			if part.body.is_multipart then
				from
					body := part.body
					i := 1
				until
					i > body.parts_count
				loop
					sub_part := body.part (i)

					-- we take care by checking if mime message is
					-- according to the specs. It should have a
					-- Content-Disposition field with value "form-data".
					-- Every body part that does not have this field, is skipped.
					-- The name parameter can be absent in case of multiple
					-- uploaded files
					sub_part.fields.search (field_name_content_disposition)
					if sub_part.fields.found and then sub_part.fields.found_item.value.is_equal ("form-data") then
						field ?= sub_part.fields.found_item
						field.parameters.search ("name")
						if field.parameters.found then
							parameter := field.parameters.found_item
						else
							parameter := Void
						end
						if parameter /= Void or else name /= Void then
							if parameter /= Void then
								keyname := parameter.value
							else
								keyname := name
							end

							-- if multipart (multiple uploaded files), recurse
							-- else singlepart with value
							if sub_part.body.is_multipart then
								do_fill_key_value_pairs_from_mime (keyname, sub_part)
							else
								field.parameters.search ("filename")
								if field.parameters.found then
									parameter := field.parameters.found_item
								else
									parameter := Void
								end
								if parameter /= Void and then not parameter.value.is_empty then
									keyvalue := remove_directory_part (parameter.value)
									file_body ?= sub_part.body
									temporary_file := file_body.file
								else
									text_body ?= sub_part.body
									keyvalue := text_body.as_string
								end
								create kv.make (keyname, keyvalue)
								if temporary_file /= Void then
									kv.set_file (temporary_file)
									temporary_file := Void
								end
								cgi_data.put (kv, kv.key)
							end
						end
					end
					i := i + 1
				end
			else -- not recognized variant... perhaps accept?
			end
		end

	fill_key_value_pairs_from_multipart_encoding is
			-- Posted data.
		local
			parser: EPX_MIME_PARSER
		do
			create parser.make_from_file (stdin)
			parser.set_header (mime_header)
			parser.parse
			if not parser.syntax_error then
				do_fill_key_value_pairs_from_mime (Void, parser.part)
			else
				raise ("Syntax error detected during parsing of POSTed data.")
			end
		end

	fill_key_value_pairs_from_standard_encoding is
			-- Read query values from url request.
		local
			mya, myb: ARRAY [STRING]
			s: STRING
			kv: EPX_KEY_VALUE
			i: INTEGER
			empty_field: BOOLEAN
			new_empty_string: STRING
		do
			set_cgi_input
			if cgi_input /= Void then
				mya := sh.split_on (cgi_input, '&')
				from
					i := mya.lower
				until
					i > mya.upper
				loop
					s := mya.item (i)
					myb := sh.split_on (s, '=')
					empty_field := myb.count = 1
					if empty_field then
						create new_empty_string.make (0)
						create kv.make (
											 decode_string (myb.item (myb.lower)),
											 new_empty_string)
					else
						create kv.make (
											 decode_string (myb.item (myb.lower)),
											 decode_string (myb.item (myb.upper)))
					end
					cgi_data.put (kv, kv.key)
					i := i + 1
				end
			end
		end

	mime_header: STRING is
			-- Suitable MIME header for POST data.
		local
			fixed_content_type: STRING
			i: INTEGER
			value_start,
			value_end: INTEGER
		once
			-- Opera contains a boundary parameter with invalid characters.
			-- Fix this by making sure the boundary parameter is properly quoted.
			fixed_content_type := clone (content_type)
			i := fixed_content_type.substring_index ("boundary=", 1)
			if i /= 0 then
				value_start := i + 9
				if value_start <= fixed_content_type.count then
					if fixed_content_type.item (value_start) /= '"' then
						STRING_.insert_character (fixed_content_type, '"', value_start)
						from
							value_end := value_start
						until
							value_end > fixed_content_type.count or else
							fixed_content_type.item (value_end) = ' ' or else
							fixed_content_type.item (value_end) = ';' or else
							fixed_content_type.item (value_end) = '%T' or else
							fixed_content_type.item (value_end) = '%N' or else
							fixed_content_type.item (value_end) = '%R'
						loop
							value_end := value_end + 1
						end
						if value_end > fixed_content_type.count then
							fixed_content_type.append_character ('"')
						else
							STRING_.insert_character (fixed_content_type, '"', value_end)
						end
					end
				end
			end
			Result := sh.make_with_capacity (field_name_content_type.count + 2 + fixed_content_type.count + 4)
			Result.append_string (field_name_content_type)
			Result.append_character (':')
			Result.append_character (' ')
			Result.append_string (fixed_content_type)
			Result.append_character ('%R')
			Result.append_character ('%N')
			Result.append_character ('%R')
			Result.append_character ('%N')
		ensure
			result_not_empty: Result /= Void and then not Result.is_empty
		end


feature {NONE} -- Microsoft Internet Explorer stuff

	remove_directory_part (filename: STRING): STRING is
			-- IE sends full path instead of just filename as everybody
			-- else does. Of course, that isn't a security hole...
		local
			path: STDC_PATH
		do
			create path.make_from_string (filename)
			path.parse (Void)
			path.set_directory (Void)
			Result := path
		end


feature {NONE} -- Decoding of encoded passed values, thanks Dustin Sallings

	decode_string (in: STRING): STRING is
			-- Decode an HTTP encoded string.
		local
			i: INTEGER
			c: CHARACTER
		do
			Result := sh.make_with_capacity (64)
			from
				i := 1
			until
				i > in.count
			loop
				c := in.item (i)
				if c = '+' then
					Result.append_character (' ')
				elseif c = '%%' then
					i := i + 1
					Result.append_character (from_hex(in.substring(i, i + 1)))
					i := i + 1
				else
					Result.append_character (c)
				end
				i := i + 1
			end
		end

	from_hex (in: STRING): CHARACTER is
			-- Take a two digit hex string, and convert it to a hex character.
		require
			has_two_digits: in.count = 2
			-- in.item(1).is_hexadecimal_digit
			-- in.item(2).is_hexadecimal_digit
		local
			mya, myb: INTEGER
			map, tmp: STRING
		do
			tmp := clone (in)
			tmp.to_upper
			map := "0123456789ABCDEF"
			mya := map.index_of (tmp.item(1), 1) - 1
			myb := map.index_of (tmp.item(2), 1) - 1
			mya := mya * 16 + myb
			Result := ch.from_integer (mya)
		end


end
